""" Contains functionality for configuration of pipeline elements """
import configargparse


class ConfigurationHelper:
    """
    Helper class for parsing pipeline element configuration from multiple sources.
    Provides a thin wrapper around the `configargparse` library.
    """

    def __init__(self):
        self.config = None
        self._parser = configargparse.ArgParser(
            default_config_files=["weasel-pipeline.conf"]
        )
        self._parser.add("-c", "--config", is_config_file=True, help="config file path")
        self._parser.add(
            "--amqp_uri",
            env_var="AMQP_URI",
            required=True,
            help="URL of RabbitMQ server",
        )

    def add_argument(
        self,
        keyword: str,
        argument_type: type = str,
        required: bool = True,
        help_text: str = None,
    ) -> None:
        """
        Add an argument to the configuration parser.

        :param keyword: The keyword for this argument, i.e. `--<keyword>` on the command line
        :param argument_type: The type of the argument value (e.g. str, int)
        :param required: Whether or not this argument is required
        :param help_text: The help text which is shown for this argument
        """
        self._parser.add(
            "--{}".format(keyword),
            env_var="{}".format(keyword.upper()),
            type=argument_type,
            required=required,
            help=help_text,
        )

    def parse(self) -> None:
        """
        Parse the passed arguments and store the result in the `config` attribute.
        """
        self.config = self._parser.parse_args()



""" Contains functionality for working with RabbitMQ """
import asyncio
import pickle

import aio_pika
from aiormq.exceptions import ChannelInvalidStateError


class QueueHelper:
    """ Helper class for exchanging messages between elements of a weasel-pipeline toolchain """

    _queues = dict()
    _connection = None
    _channel = None

    @staticmethod
    async def create(amqp_uri: str, prefetch_count: int = 5000):
        """
        Create a QueueHelper instance and establish connection to the RabbitMQ server.

        :param amqp_uri: The connection URI for the RabbitMQ server
        :param prefetch_count:
            Controls the number of messages which are pre-fetched from each queue.
            Keeping this number high enough will significantly improve message throughput
        :return: The created QueueHelper instance
        """
        queue_helper = QueueHelper()
        await queue_helper.initialize_pika(amqp_uri, prefetch_count)
        return queue_helper

    async def initialize_pika(self, amqp_uri: str, prefetch_count: int) -> None:
        """
        Initialize the connection to RabbitMQ.
        This is only required if the instance is not created via the `create` method.

        :param amqp_uri: The connection URI for the RabbitMQ server
        :param prefetch_count:
            Controls the number of messages which are pre-fetched from each queue.
            Keeping this number high enough will significantly improve message throughput
        """
        self._connection = await aio_pika.connect_robust(
            amqp_uri, loop=asyncio.get_event_loop()
        )
        self._channel = await self._connection.channel()
        await self._channel.set_qos(prefetch_count=prefetch_count)

    async def close(self) -> None:
        """
        Close the connection to RabbitMQ.
        """
        await self._channel.close()
        await self._connection.close()

    async def declare_queue(self, name: str, queue_max_length: int = 1000) -> None:
        """
        Declare a queue with the given name.

        :param name: The identifying name for the created queue
        :param queue_max_length:
            Defines the maximum length of this queue. This prevents the queue from growing
            indefinitely. When an element calls `put_value` to a queue which is currently full,
            this call will be suspended until there is free space in the queue. This way all
            elements in the pipeline get restricted to the speed of the bottleneck elements.
        """
        queue = await self._channel.declare_queue(
            name,
            arguments={
                "x-max-length": queue_max_length,
                "x-overflow": "reject-publish",
            },
        )
        self._queues[name] = queue

    async def put_value(self, name: str, value: object) -> None:
        """
        Output a value to the queue of the given name. If the queue is currently full, this
        call will be suspended until there is free space again.

        :param name: The identifying name of the queue to which the value is written
        :param value: The output value. Can be any data type supported by `pickle`
        """
        while True:
            try:
                await self._channel.default_exchange.publish(
                    aio_pika.Message(body=pickle.dumps(value)), routing_key=name
                )
                return
            except (
                aio_pika.exceptions.DeliveryError,
                ChannelInvalidStateError,
                ConnectionResetError,
                ConnectionRefusedError,
                RuntimeError,
            ):
                await asyncio.sleep(0.1)

    async def pull_values(self, name: str):
        """
        Pull values from the queue of the given name. This will return an iterator that is
        supposed to be used in an `async for` loop.

        :param name: The identifying name of the queue from which the values are read
        """
        async for message in self._queues[name]:
            async with message.process():
                yield pickle.loads(message.body)



""" Contains functionality for working with asyncio """
import asyncio

import uvloop


class AsyncContext:
    """ Helper class which installs uvloop to the asyncio event loop """

    _loop = None

    @classmethod
    def get_loop(cls) -> asyncio.AbstractEventLoop:
        """
        Get the event loop. This class method always returns the same event loop,
        realizing a singleton pattern, to avoid conflicting event loop objects. It
        also installs uvloop to the event loop instance to maximize performance.
        """
        if cls._loop is None:
            uvloop.install()
            cls._loop = asyncio.get_event_loop()
        return cls._loop

    @classmethod
    def run(cls, async_fn) -> None:
        """
        Run async code in the uvloop-patched event loop.

        :param async_fn: The awaitable function/method to run
        """
        cls.get_loop().run_until_complete(async_fn)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['weasel_pipeline']

package_data = \
{'': ['*']}

install_requires = \
['aio-pika>=6.7.1,<7.0.0',
 'configargparse>=1.2.3,<2.0.0',
 'uvloop>=0.14.0,<0.15.0']

setup_kwargs = {
    'name': 'weasel-pipeline',
    'version': '1.1.7',
    'description': 'A minimalist pipelining framework for the WEASEL project.',
    'long_description': 'WEASEL Pipeline\n===============\n\n[![pipeline status](https://gitlab.com/weasel-project/weasel-pipeline/badges/master/pipeline.svg)](https://gitlab.com/weasel-project/weasel-pipeline/-/commits/master)\n[![PyPI version](https://img.shields.io/pypi/v/weasel-pipeline)](https://pypi.org/project/weasel-pipeline/)\n[![Documentation](https://readthedocs.org/projects/weasel-pipeline/badge/?version=latest&style=flat)](https://weasel-pipeline.readthedocs.io/en/latest/)\n[![dependency status](https://img.shields.io/librariesio/release/pypi/weasel-pipeline)](https://libraries.io/pypi/weasel-pipeline)\n![PyPI license](https://img.shields.io/pypi/l/weasel-pipeline)\n![PyPI Python version](https://img.shields.io/pypi/pyversions/weasel-pipeline)\n\n`weasel-pipeline` is a minimalist pipelining framework that serves as a common foundation of our scanning and data processing tool chains.\n\nInstallation instructions, examples, and further documentation available [here](https://weasel-pipeline.readthedocs.io/en/latest/).',
    'author': 'Fabian Marquardt',
    'author_email': 'marquard@cs.uni-bonn.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/weasel-project/weasel-pipeline',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
